#define _WSS_DEBUG

#ifndef _WSS_DEBUG
#endif

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <go32.h>
#include <dpmi.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <sys/farptr.h>
#include <pc.h>
#include <ctype.h>
#include <stdarg.h>
#include "wss.h"

  static void logerror_(const char *text,...)
  {
	  va_list arg;

	  va_start(arg,text);
	  vfprintf(stdout,text,arg);
	  va_end(arg);
  }

  /***********************	TICKER	*************************/

  __extension__ typedef signed long long	  INT64;
  __extension__ typedef unsigned long long	  UINT64;
  typedef INT64 cycles_t;


  cycles_t osd_cycles(void)
  {
  #if 0
	  INT64 result;

	  __asm__ __volatile__ (
		  "rdtsc"
		  : "=A" (result)
	  );

	  return result;
  #else
	  return uclock();
  #endif
  }

  #define osd_cycles_per_second() UCLOCKS_PER_SEC


/***********************  COMMON  *************************/


#define _8BITMONO		1
#define _8BITSTEREO 	2
#define _16BITSTEREO	3


typedef struct {
	int 	initialized;
	void	(*device_exit)(void);
	void	(*pcm_upload)(void);
	DWORD	(*get_current_pos)(void);
	DWORD	playback_rate;
	int 	pcm_format;
	char	*device_name;
	int 	irq;
	int 	isa_port;
	int 	isa_dma;
	int 	isa_hdma;
} WAVEDEVICE;

static WAVEDEVICE wd = { FALSE, NULL, NULL, NULL, 0, _16BITSTEREO, NULL, -1, -1, -1, -1};

static void wavedevice_struct_init(void)
{
	wd.initialized = FALSE;
	wd.device_exit = NULL;
	wd.pcm_upload  = NULL;
	wd.get_current_pos = NULL;
	wd.playback_rate = 0;
	wd.pcm_format  = _16BITSTEREO;
	wd.device_name = NULL;
	wd.irq		= -1;
	wd.isa_port = -1;
	wd.isa_dma	= -1;
	wd.isa_hdma = -1;
}


static void common_pcm_upload_func(void);
static DWORD common_dma_current_pos(void);



static BOOL _dma_allocate_mem4k(int *sel, unsigned long *phys);
static int _dma_allocate_mem(int *sel, unsigned long *phys);
static void _dma_start(int channel, unsigned long addr, int size, int auto_init, int input);
static void _dma_stop(int channel);
static DWORD _dma_todo(int channel);
static int _dma_count;
static DWORD _dma_counter(void);

#define SAMPLECNT	  16384 		// the buffer size has to be smaller than 64KBytes, for compatibility sake.
#define SAMPLECNTMASK 0x3FFF		// 16bitPCM * stereo * 16384 = 64KBytes


static DWORD g_prev_play_cursor = 0;
static DWORD g_write_cursor = 0;
static DWORD g_latency	= 48000/30;
static float g_latencym = 2.2;
static DWORD g_dma_average_cnt = (48000/60) << 8;
static int	 g_dma_dt = 0;
static int	 g_dma_overflow  = 0;
static int	 g_dma_underflow = 0;

#define DMA_AVERAGE 			256
#define DMA_AVERAGE_MASK		0xFF
#define DMA_AVERAGE_SHIFT_COUNT 8

static DWORD g_dma_remainder = 0;
static long  mixing_buff[SAMPLECNT*2];
static DWORD g_current_req = 0;
static DWORD g_next_req = 0;
static int	 g_master_volume = 256;
static DWORD g_samples_per_frame = 0;



/* dos memory 4k bytes for work area */
static BOOL allocate_dosmem4k(void);
static void free_dosmem4k(void);
static DWORD get_address_dosmem4k(void);
static int g_dosmem4k_sel = NULL;
static unsigned long g_dosmem4k_addr = NULL;
static DWORD g_dosmem4k_phys_table[1];
//static __dpmi_meminfo info4k;

/* dos memory 64k cyclic playing buffer */
static BOOL allocate_dosmem64k_for_dma(int format);
static void free_dosmem64k_for_dma(void);
static DWORD get_address_dosmem64k_for_dma(void);
static int g_wss_dma_sel = NULL;
static unsigned long g_wss_dma_addr = NULL;
static DWORD g_dosmem64k_phys_table[16];
static int g_dma_buff_size;
static int g_dmacnt_shift_count;
static __dpmi_meminfo info64k;

static int tick = 0;

static DWORD actual_sample_rate = 0;
static int sound_device_master_volume  = -1;	// 0:max - 31:min, or -1(default)

static int stereo_reverse_flag = FALSE;


/*	function  */

static DWORD get_played_sample_count(void);
static void calc_next_req(void);


/* error message func */

#define ERROR_MESSAGE_BUFFER_SIZE	1024
static int error_message_pointer = 0;
static char error_message[ERROR_MESSAGE_BUFFER_SIZE] = "\n";

static void set_error_message(char *message)
{
	while(1){
		if(error_message_pointer >= (ERROR_MESSAGE_BUFFER_SIZE - 1)) break;
		if(*message == '\0') break;
		error_message[error_message_pointer++] = *message++;
	}
	error_message[error_message_pointer] = '\0';
}

static char* get_error_message(void)
{
	return &error_message[0];
}

static void clear_error_message(void)
{
	error_message[0] = '\0';
	error_message_pointer = 0;
}


/***********************  COMMON END  *********************/














/********************************************************************
	PCI BIOS helper funtions
 ********************************************************************/

#define PCI_ANY_ID	((WORD)(~0))

typedef struct {
	WORD vender_id;
	WORD device_id;
	WORD sub_vender_id;
	WORD sub_device_id;
	WORD device_bus_number;
	DWORD base0;
	DWORD base1;
	DWORD base2;
	DWORD base3;
	DWORD base4;
	DWORD base5;
} PCI_DEV;

static PCI_DEV g_pci;


static BOOL pci_read_config_byte(PCI_DEV *pci, int idx, BYTE *data)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B108;						/* read config byte */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.d.edi = (DWORD)idx;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("pci read config byte failed\n");
		result = FALSE;
		r.d.ecx = 0;
	}
	*data = (BYTE)r.d.ecx;
	return result;
}

static BOOL pci_read_config_word(PCI_DEV *pci, int idx, WORD *data)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B109;						/* read config word */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.d.edi = (DWORD)idx;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("pci read config word failed\n");
		result = FALSE;
		r.d.ecx = 0;
	}
	*data = (WORD)r.d.ecx;
	return result;
}

static BOOL pci_read_config_dword(PCI_DEV *pci, int idx, DWORD *data)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B10A;						/* read config dword */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.d.edi = (DWORD)idx;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("pci read config dword failed\n");
		result = FALSE;
		r.d.ecx = 0;
	}
	*data = (DWORD)r.d.ecx;
	return result;
}

static BOOL pci_write_config_byte(PCI_DEV *pci, int idx, BYTE data)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B10B;						/* write config byte */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.d.ecx = (DWORD)data;
	r.d.edi = (DWORD)idx;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("pci write config byte failed\n");
		result = FALSE;
	}
	return result;
}

static BOOL pci_write_config_word(PCI_DEV *pci, int idx, WORD data)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B10C;						/* write config word */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.d.ecx = (DWORD)data;
	r.d.edi = (DWORD)idx;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("pci write config word failed\n");
		result = FALSE;
	}
	return result;
}

static BOOL pci_write_config_dword(PCI_DEV *pci, int idx, DWORD data)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B10D;						/* write config dword */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.d.ecx = (DWORD)data;
	r.d.edi = (DWORD)idx;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("pci write config dword failed\n");
		result = FALSE;
	}
	return result;
}

static BOOL set_pci_hardware_interrupt(PCI_DEV *pci, BYTE int_pin, BYTE irq)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B10F;						/* set pci hardware interrupt */
	r.d.ebx = (DWORD)pci->device_bus_number;
	r.h.cl	= int_pin;
	r.h.ch	= irq;
	r.x.ds	= 0xF000;
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		logerror_("set pci hardware interrupt failed.\n");
		result = FALSE;
	}
	pci_write_config_byte(pci, 0x3C, irq);
	return result;
}


static BOOL check_pci_bios(void)
{
	__dpmi_regs r;
	BOOL result = TRUE;

	r.d.eax = 0x0000B101;					// PCI BIOS - INSTALLATION CHECK
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if( r.d.edx != 0x20494350 ){			// ' ICP'
		result = FALSE;
	}
	return result;
}


static BOOL find_pci_device(PCI_DEV *pci)
{
	__dpmi_regs r;
	WORD wdata;

	r.d.eax = 0x0000B102;					// PCI BIOS - FIND PCI DEVICE
	r.d.ecx = pci->device_id;				// device ID
	r.d.edx = pci->vender_id;				// vendor ID
	r.d.esi = 0x00000000;					// device index
	__dpmi_int(0x1a, &r);
	if( r.h.ah != 0 ){
		return FALSE;						// no specified device found
	}
	pci->device_bus_number = r.x.bx;		// save the device & bus number
	if(pci->sub_vender_id != PCI_ANY_ID){
		/* get subsystem vender id */
		if(pci_read_config_word(pci, 0x2C, &wdata) == FALSE) return FALSE;
		if(wdata != pci->sub_vender_id) return FALSE;
	}
	if(pci->sub_device_id != PCI_ANY_ID){
		/* get subsystem device id */
		if(pci_read_config_word(pci, 0x2E, &wdata) == FALSE) return FALSE;
		if(wdata != pci->sub_device_id) return FALSE;
	}
	/* device found */
	return TRUE;
}


static void enable_pci_io_access(PCI_DEV *pci)
{
	WORD wdata;

	pci_read_config_word(pci, 0x04, &wdata);
	/*	set I/O access enable bit  */
	wdata |= 0x0001;
	pci_write_config_word(pci, 0x04, wdata);
}


static void enable_pci_memory_access(PCI_DEV *pci)
{
	WORD wdata;

	pci_read_config_word(pci, 0x04, &wdata);
	/*	set memory access enable bit  */
	wdata |= 0x0002;
	pci_write_config_word(pci, 0x04, wdata);
}

static void enable_pci_bus_master(PCI_DEV *pci)
{
	WORD wdata;

	pci_read_config_word(pci, 0x04, &wdata);
	/*	set bus master enable bit  */
	wdata |= 0x0004;
	pci_write_config_word(pci, 0x04, wdata);
}


/********************************************************************
 *	Windows detection
 ********************************************************************/

static BOOL detect_windows(void)
{
	// return  TRUE:	windows detected
	//		  FALSE: no windows detected

	__dpmi_regs r;

	r.x.ax = 0x1600;
	__dpmi_int(0x2F, &r);
//	  printf("ax = %04x\n", r.x.ax);
	if( (r.h.al & 0x7F) != 0 ){
		return TRUE;
	}
	return FALSE;
}


/********************************************************************
	Interrupt and others helper functions
 ********************************************************************/

static DWORD tasmania = 0;

void w_enter_critical(void)
{
	__dpmi_regs r;

	tasmania += 1;
	if(tasmania == 1){
		r.x.ax = 0x1681;
		__dpmi_int(0x2F, &r);
		asm volatile("cli");
	}
}

void w_exit_critical(void)
{
	__dpmi_regs r;

	tasmania -= 1;
	if(tasmania == 0){
		r.x.ax = 0x1682;
		__dpmi_int(0x2F, &r);
		asm volatile("sti");
	}
}

static void eoi(int irq)
{
	outportb(0x20, 0x20);
	if (irq > 7) outportb(0xA0, 0x20);
}

static void unmask_irq(int irq)
{
	BYTE d0;

	d0 = inportb(0x21);

	if (irq > 7) {
		outportb(0x21, d0 & 0xFB);
		d0 = inportb(0xA1);
		outportb(0xA1, d0 & ~(1<<(irq-8)) );
   }else{
	  outportb(0x21, d0 & (~(1<<irq)) );
   }
}

static void mask_irq(int irq)
{
	BYTE d0;

	if (irq > 7) {
		d0 = inportb(0xA1);
		outportb(0xA1, d0 | (1<<(irq-8)));
	}else{
		d0 = inportb(0x21);
		outportb(0x21, d0 | (1<<irq));
	}
}

static void udelay(int usec)
{
	cycles_t  prev;
	double t;
	cycles_t  cps = osd_cycles_per_second();

	t = ((double)cps / 1000000.0) * usec;
	prev = osd_cycles();
	while(1){
		if( (osd_cycles() - prev) >= (cycles_t)t ) break;
	}
}



/********************************************************************
	DOS memory allocation helper funtions
 ********************************************************************/

static void copy_to_dos_memory(DWORD dos_address, BYTE *src, DWORD length)
{
	int d0;

	_farsetsel(_dos_ds);
	d0 = 0;
	while(d0 < length){
		_farnspokeb(dos_address + d0, *(src + d0));
		d0 += 1;
	}
}

static void copy_from_dos_memory(DWORD dos_address, BYTE *dest, DWORD length)
{
	int d0;

	_farsetsel(_dos_ds);
	d0 = 0;
	while(d0 < length){
		*(dest + d0) = _farnspeekb(dos_address + d0);
		d0 += 1;
	}
}


static BOOL allocate_dosmem4k(void)
{
	if(g_dosmem4k_sel  != NULL) return FALSE;
	if(g_dosmem4k_addr != NULL) return FALSE;
	if(_dma_allocate_mem4k(&g_dosmem4k_sel, &g_dosmem4k_addr) == FALSE){
		return FALSE;
	}
//	  info4k.size	 = 4096;
//	  info4k.address = g_dosmem4k_addr;
//	  __dpmi_lock_linear_region(&info4k);

	return TRUE;
}

static void free_dosmem4k(void)
{
	if(g_dosmem4k_sel  == NULL) return;
	if(g_dosmem4k_addr == NULL) return;
//	  if(info4k.address != NULL){
//		  __dpmi_unlock_linear_region(&info4k);
//	  }
	__dpmi_free_dos_memory(g_dosmem4k_sel);
}

static DWORD get_address_dosmem4k(void)
{
	return g_dosmem4k_addr;
}


static BOOL allocate_dosmem64k_for_dma(int format)
{
	BYTE *a0;
	int d0;
	BOOL lockflag = FALSE;

	g_dma_buff_size = SAMPLECNT;
	switch(format){
		case _8BITMONO:
			g_dma_buff_size *= 1;
			g_dmacnt_shift_count = 0;
			break;
		case _8BITSTEREO:
			g_dma_buff_size *= 2;
			g_dmacnt_shift_count = 1;
			break;
		case _16BITSTEREO:
			g_dma_buff_size *= 4;
			g_dmacnt_shift_count = 2;
			break;
		default:
			g_dma_buff_size = 0;
	}
	if(g_dma_buff_size == 0) return FALSE;
	a0 = malloc(g_dma_buff_size);
	if(a0 == NULL) return FALSE;
	if (_dma_allocate_mem(&g_wss_dma_sel, &g_wss_dma_addr) == FALSE)
		return FALSE;
	d0 = 0;
	while(d0 < g_dma_buff_size){
		if(format == _8BITSTEREO || format == _8BITMONO)
			a0[d0] = 0x80;
		else
			a0[d0] = 0x00;
		d0 += 1;
	}
	copy_to_dos_memory(g_wss_dma_addr, a0, g_dma_buff_size);
	free(a0);

	if(lockflag == TRUE){
		info64k.size	= 65536;
		info64k.address = g_wss_dma_addr;
		__dpmi_lock_linear_region(&info64k);
	}else{
		info64k.address = NULL;
	}

	return TRUE;
}

static void free_dosmem64k_for_dma(void)
{
	if(g_wss_dma_sel  == NULL) return;
	if(g_wss_dma_addr == NULL) return;
	if(info64k.address != NULL){
		__dpmi_unlock_linear_region(&info64k);
	}
	__dpmi_free_dos_memory(g_wss_dma_sel);
}

static DWORD get_address_dosmem64k_for_dma(void)
{
	return g_wss_dma_addr;
}





/********************************************************************
	VDS helper functions
 ********************************************************************/

typedef struct {
	DWORD Region_Size		__attribute__ ((packed));
	DWORD Offset			__attribute__ ((packed));
	WORD  Seg_or_Select 	__attribute__ ((packed));
	WORD  Buffer_ID 		__attribute__ ((packed));
	DWORD Physical_Address	__attribute__ ((packed));
} DDS;


static BOOL is_vds_available(void)
{
	__dpmi_regs r;
	int available;
	BOOL flag = FALSE;

	available = (_farpeekb (_dos_ds, 0x0047b) & 0x20) != 0;

	if(available != 0){
		// get vds verson nunber
		r.x.ax = 0x8102;
		r.x.dx = 0x0000;
		__dpmi_int(0x4B, &r);
		if((r.x.flags & 0x01) == 0){		// carry is clear if successful
			if(r.x.ax != 0x8102){
				flag = TRUE;
			}
		}
	}

	return flag;
}


static BOOL vds_helper_lock(DWORD buffer_addr, DWORD *physical_addr_table, int pages)
{
	__dpmi_regs r;
	DWORD temp;
	DDS dds;
	int d0;
	int sel;
	unsigned long dos_addr;
	BOOL vds_available;

	vds_available = is_vds_available();

	if(_dma_allocate_mem4k(&sel, &dos_addr) == FALSE){
		set_error_message("vds_helper_lock: _dma_allocate_mem4k() error.\n");
		return FALSE;
	}
	memset(&dds, 0, sizeof(dds));

	d0 = 0;
	while(d0 < pages){
		temp = buffer_addr + (4096 * d0);
		if(vds_available == TRUE){
			/* vds available */
			dds.Region_Size   = 4096;
			dds.Offset		  = temp & 0x0F;
			dds.Seg_or_Select = temp >> 4;
//			printf("%08x : %08x : %08x\n", temp, dds.Offset, dds.Seg_or_Select);
			copy_to_dos_memory(dos_addr, (BYTE*)&dds, sizeof(dds));
			r.x.ax = 0x8103;			// LOCK DMA BUFFER REGION
			r.x.dx = 0x0000;
			r.x.di = dos_addr & 0xF;
			r.x.es = dos_addr >> 4;
			__dpmi_int(0x4B, &r);
			if((r.x.flags & 0x01) != 0){
				set_error_message("vds_helper_lock: LOCK DMA BUFFER REGION error\n");
				__dpmi_free_dos_memory(sel);
				return FALSE;
			}
			copy_from_dos_memory(dos_addr, (BYTE*)&dds, sizeof(dds));
			if(dds.Buffer_ID != 0){
				set_error_message("vds_helper_lock: Buffer_ID allocated. error\n");
				__dpmi_free_dos_memory(sel);
				return FALSE;
			}
			physical_addr_table[d0] = dds.Physical_Address;
		}else{
			/* vds not available */
			physical_addr_table[d0] = temp;
		}
		d0 += 1;
	}
	__dpmi_free_dos_memory(sel);

	return TRUE;
}

static BOOL vds_helper_unlock(DWORD buffer_addr, DWORD *physical_addr_table, int pages)
{
	__dpmi_regs r;
	DWORD temp;
	DDS dds;
	int d0;
	int sel;
	unsigned long dos_addr;
	BOOL vds_available;

	vds_available = is_vds_available();

	if(_dma_allocate_mem4k(&sel, &dos_addr) == FALSE){
		set_error_message("vds_helper_unlock: _dma_allocate_mem4k() error.\n");
		return FALSE;
	}
	memset(&dds, 0, sizeof(dds));

	d0 = 0;
	while(d0 < pages){
		if(vds_available == TRUE){
			/* vds available */
			temp = buffer_addr + (4096 * d0);
			dds.Region_Size   = 4096;
			dds.Offset		  = temp & 0x0F;
			dds.Seg_or_Select = temp >> 4;
			dds.Buffer_ID	  = 0;
			dds.Physical_Address = physical_addr_table[d0];
//			printf("%08x : %08x : %08x\n", temp, dds.Offset, dds.Seg_or_Select);
			copy_to_dos_memory(dos_addr, (BYTE*)&dds, sizeof(dds));
			r.x.ax = 0x8104;			// UNLOCK DMA BUFFER REGION
			r.x.dx = 0x0000;
			r.x.di = dos_addr & 0xF;
			r.x.es = dos_addr >> 4;
			__dpmi_int(0x4B, &r);
			if((r.x.flags & 0x01) != 0){
				set_error_message("vds_helper_unlock: UNLOCK DMA BUFFER REGION error\n");
				__dpmi_free_dos_memory(sel);
				return FALSE;
			}
		}
		physical_addr_table[d0] = NULL;
		d0 += 1;
	}
	__dpmi_free_dos_memory(sel);

	return TRUE;
}


/********************************************************************
	DMA helper functions
	This is based on Allegro liblary.
 ********************************************************************/

// 8bit DMAC
#define DMA1_STAT		0x08
#define DMA1_WCMD		0x08
#define DMA1_WREQ		0x09
#define DMA1_SNGL		0x0A
#define DMA1_MODE		0x0B
#define DMA1_CLRFF		0x0C
#define DMA1_MCLR		0x0D
#define DMA1_CLRM		0x0E
#define DMA1_WRTALL 	0x0F

// 16bit DMAC
#define DMA2_STAT		0xD0
#define DMA2_WCMD		0xD0
#define DMA2_WREQ		0xD2
#define DMA2_SNGL		0xD4
#define DMA2_MODE		0xD6
#define DMA2_CLRFF		0xD8
#define DMA2_MCLR		0xDA
#define DMA2_CLRM		0xDC
#define DMA2_WRTALL 	0xDE


#define DMA0_ADDR		0x00
#define DMA0_CNT		0x01
#define DMA1_ADDR		0x02
#define DMA1_CNT		0x03
#define DMA2_ADDR		0x04
#define DMA2_CNT		0x05
#define DMA3_ADDR		0x06
#define DMA3_CNT		0x07
#define DMA4_ADDR		0xC0
#define DMA4_CNT		0xC2
#define DMA5_ADDR		0xC4
#define DMA5_CNT		0xC6
#define DMA6_ADDR		0xC8
#define DMA6_CNT		0xCA
#define DMA7_ADDR		0xCC
#define DMA7_CNT		0xCE

#define DMA0_PAGE		0x87
#define DMA1_PAGE		0x83
#define DMA2_PAGE		0x81
#define DMA3_PAGE		0x82
#define DMA4_PAGE		0x8F
#define DMA5_PAGE		0x8B
#define DMA6_PAGE		0x89
#define DMA7_PAGE		0x8A


typedef struct {
   BYTE dma_disable;
   BYTE dma_enable;
   WORD page;
   WORD addr;
   WORD count;
   WORD single;
   WORD mode;
   WORD clear_ff;
   BYTE write;
   BYTE read;
} DMA_S;



static DMA_S mydma[] =
{
   /* channel 0 */
   { 0x04, 0x00, DMA0_PAGE, DMA0_ADDR, DMA0_CNT,
	 DMA1_SNGL, DMA1_MODE, DMA1_CLRFF, 0x48, 0x44 },

   /* channel 1 */
   { 0x05, 0x01, DMA1_PAGE, DMA1_ADDR, DMA1_CNT,
	 DMA1_SNGL,DMA1_MODE,DMA1_CLRFF,0x49,0x45 },

   /* channel 2 */
   { 0x06, 0x02, DMA2_PAGE, DMA2_ADDR, DMA2_CNT,
	 DMA1_SNGL, DMA1_MODE, DMA1_CLRFF, 0x4A, 0x46 },

   /* channel 3 */
   { 0x07, 0x03, DMA3_PAGE, DMA3_ADDR, DMA3_CNT,
	 DMA1_SNGL, DMA1_MODE, DMA1_CLRFF, 0x4B, 0x47 },

   /* channel 4 */
   { 0x04, 0x00, DMA4_PAGE, DMA4_ADDR, DMA4_CNT,
	 DMA2_SNGL, DMA2_MODE, DMA2_CLRFF, 0x48, 0x44 },

   /* channel 5 */
   { 0x05, 0x01, DMA5_PAGE, DMA5_ADDR, DMA5_CNT,
	 DMA2_SNGL, DMA2_MODE, DMA2_CLRFF, 0x49, 0x45 },

   /* channel 6 */
   { 0x06, 0x02, DMA6_PAGE, DMA6_ADDR, DMA6_CNT,
	 DMA2_SNGL, DMA2_MODE, DMA2_CLRFF, 0x4A, 0x46 },

   /* channel 7 */
   { 0x07, 0x03, DMA7_PAGE, DMA7_ADDR, DMA7_CNT,
	 DMA2_SNGL, DMA2_MODE, DMA2_CLRFF, 0x4B, 0x47 }
};



__inline__ BYTE get_8bit_pcm_value(long value)
{
	BYTE d0;

	if(value >	32767 - 128) value =  32767 - 128;
	if(value < -32768)		 value = -32768;
	d0 = (value + 128 - ((value >> 15) & 0x0001)) >> 8;

	return (d0 + 0x80);
}

#if 0

static int _dma_allocate_mem(int bytes, int *sel, unsigned long *phys)
{
   int seg;

   /* allocate twice as much memory as we really need */
   seg = __dpmi_allocate_dos_memory((bytes*2 + 15) >> 4, sel);

   if (seg < 0) {
	  *sel = 0;
	  *phys = 0;
	  return -1;
   }

   *phys = seg << 4;

   /* if it crosses a page boundary, use the second half of the block */
   if ((*phys>>16) != ((*phys+bytes)>>16))
	  *phys += bytes;

   return 0;
}

#endif

static int _dma_allocate_mem(int *sel, unsigned long *phys)
{
	int seg;

	/* allocate twice as much memory as we really need */
	seg = __dpmi_allocate_dos_memory(131072 >> 4, sel); 	// 128K

	if (seg < 0) {
	   *sel = 0;
	   *phys = 0;
	   return FALSE;
	}

   *phys = seg << 4;

   *phys = (*phys + 65535) & ~0xFFFF ;						// 64K boundary

   return TRUE;
}

static BOOL _dma_allocate_mem4k(int *sel, unsigned long *phys)
{
	int seg;

	/* allocate twice as much memory as we really need */
	seg = __dpmi_allocate_dos_memory(8192 >> 4, sel);	  // 8K

	if (seg < 0) {
	   *sel = 0;
	   *phys = 0;
	   return FALSE;
	}

   *phys = seg << 4;

   *phys = (*phys + 4095) & ~0x0FFF ;					   // 4K boundary

   return TRUE;
}



static void _dma_start(int channel, unsigned long addr, int size, int auto_init, int input)
{
   DMA_S *tdma;
   unsigned long page, offset;
   int mode;

   tdma = &mydma[channel];
   page = addr >> 16;

   if (channel >= 4) {
	  addr >>= 1;
	  size >>= 1;
   }

   offset = addr & 0xFFFF;
   size--;

   _dma_count = size;

   mode = (input) ? tdma->read : tdma->write;
   if (auto_init)
	  mode |= 0x10;

   outportb(tdma->single, tdma->dma_disable);
   outportb(tdma->mode, mode);
   outportb(tdma->clear_ff, 0);
   outportb(tdma->addr, offset & 0xFF);
   outportb(tdma->addr, offset >> 8);
   outportb(tdma->page, page);
   outportb(tdma->clear_ff, 0);
   outportb(tdma->count, size & 0xFF);
   outportb(tdma->count, size >> 8);
   outportb(tdma->single, tdma->dma_enable);
}


static void _dma_stop(int channel)
{
   DMA_S *tdma = &mydma[channel];
   outportb(tdma->single, tdma->dma_disable);
}







static DWORD _dma_todo(int channel)
{
	int val1, val2 = 0;
	DMA_S *tdma = &mydma[channel];
	int _dma_todo_limit = 0;

	outportb(tdma->clear_ff, 0xff);
	while(_dma_todo_limit < 1024){
		val1  = inportb(tdma->count);
		val1 |= inportb(tdma->count) << 8;
		val2  = inportb(tdma->count);
		val2 |= inportb(tdma->count) << 8;
		val1 -= val2;
		if(0 <= val1 && val1 < 8) break;
		_dma_todo_limit += 1;
	}
	if(val2 > _dma_count) val2 = _dma_count;
	if(channel > 3) val2 <<= 1;
	return val2;
}



/********************************************************************
	AC'97 drivers
	This is based on ALSA drivers
 ********************************************************************/

#define AC97_RESET				0x00	/* Reset */
#define AC97_MASTER 			0x02	/* Master Volume */
#define AC97_HEADPHONE			0x04	/* Headphone Volume (optional) */
#define AC97_PCM				0x18	/* PCM Volume */
#define AC97_EXTENDED_ID		0x28	/* Extended Audio ID */
#define AC97_EXTENDED_STATUS	0x2a	/* Extended Audio Status */
#define AC97_PCM_FRONT_DAC_RATE 0x2c	/* PCM Front DAC Rate */

#define AC97_EA_VRA 	0x0001	/* Variable bit rate enable bit */


typedef struct {
	WORD vender_id;
	WORD device_id;
	WORD sub_vender_id;
	WORD sub_device_id;
	int  type;
	char *string;
} AC97_DEVICE_LIST;


/********************************************************************
	VIA 686/8233 driver

 note:
	This driver may not work with 8233 yet.
 ********************************************************************/


#define VIA686_AC97ControllerCommand   0x80

#define VIA686_AC97ControllerBusy	   (1 << 24)
#define VIA686_PrimaryCodecDataValid   (1 << 25)

#define VIA686_CodecRead   (1 << 23)
#define VIA686_CodecWrite  0
#define VIA686_INDEX_MASK  0x7F

#define VIA686_PrimaryCodec    0x00000000

#define VIA_REG_OFFSET_STATUS		0x00	/* byte - channel status */
#define VIA_REG_OFFSET_TYPE 		0x02	/* byte - channel type */
#define VIA_REG_OFFSET_CURR_PTR 	0x04	/* dword - channel current pointer */
#define VIA_REG_PLAYBACK_CURR_COUNT 0x0C	/* dword - channel current count */

#define VIA_REG_OFFSET_CONTROL		0x01	/* byte - channel control */
#define   VIA_REG_CTRL_START		0x80	/* WO */
#define   VIA_REG_CTRL_TERMINATE	0x40	/* WO */
#define   VIA_REG_CTRL_PAUSE		0x08	/* RW */
#define   VIA_REG_CTRL_RESET		0x01	/* RW - probably reset? undocumented */
// VT8233
#define   VIA_REG_CTRL_AUTOSTART	0x20

// VT8233
#define VIA_REG_OFFSET_STOP_IDX 	0x08	/* dword - stop index, channel type, sample rate */
#define   VIA_REG_TYPE_16BIT		0x00200000	/* RW */
#define   VIA_REG_TYPE_STEREO		0x00100000	/* RW */
#define VIA_REG_OFFSET_CURR_INDEX	0x0f	/* byte - channel current index */


#define DEVICE_VT82C686    0
#define DEVICE_VT8233	   1

static AC97_DEVICE_LIST via686_dev_list[] = {
	{ 0x1106, 0x3058, PCI_ANY_ID, PCI_ANY_ID, DEVICE_VT82C686, "VIA VT82C686 integrated AC97 audio" },
	{ 0x1106, 0x3059, PCI_ANY_ID, PCI_ANY_ID, DEVICE_VT8233,   "VIA VT8233 integrated AC97 audio" },

	{ 0x0000, 0x0000, PCI_ANY_ID, PCI_ANY_ID, 0, "" }
};

static int via686_device_type = DEVICE_VT82C686;


static DWORD via686_ReadAC97Codec_sub(void)
{
	DWORD d0;
	int limit = 1024;

	/* wait for ready */
	do{
		d0 = inportl(g_pci.base0 + VIA686_AC97ControllerCommand);
//		  printf("read debug %x\n", d0);
		/* always valid, maybe */
		if( (d0 & VIA686_PrimaryCodecDataValid) != 0 ) break;
		udelay(100);
	}while(limit-- > 0);

	d0 = inportl(g_pci.base0 + VIA686_AC97ControllerCommand);
	return d0;
}


static void via686_WriteAC97Codec_sub(DWORD value_to_write)
{
	DWORD d0;
	int limit = 1024;

	/* wait for ready (writing) */
	do{
		d0 = inportl(g_pci.base0 + VIA686_AC97ControllerCommand);
//		  printf("%x\n", d0);
		if( (d0 & VIA686_AC97ControllerBusy) == 0 ) break;
		udelay(100);
	}while(limit-- > 0);

	outportl(g_pci.base0 + VIA686_AC97ControllerCommand, value_to_write);

	/* wait for ready (reading) */
	limit = 1024;
	do{
		d0 = inportl(g_pci.base0 + VIA686_AC97ControllerCommand);
//		  printf("%x\n", d0);
		if( (d0 & VIA686_AC97ControllerBusy) == 0 ) break;
		udelay(100);
	}while(limit-- > 0);
}

static WORD via686_ReadAC97Codec(BYTE idx)
{
	DWORD d0;

	idx = idx & VIA686_INDEX_MASK;
	d0 = VIA686_PrimaryCodec | VIA686_CodecRead | ((DWORD)idx << 16);
	via686_WriteAC97Codec_sub(d0);
	via686_WriteAC97Codec_sub(d0);

	return via686_ReadAC97Codec_sub();
}

static void via686_WriteAC97Codec(BYTE idx, WORD data)
{
	DWORD d0;

	idx = idx & VIA686_INDEX_MASK;
	d0 = VIA686_PrimaryCodec | VIA686_CodecWrite | ((DWORD)idx << 16) | data;
	via686_WriteAC97Codec_sub(d0);
	via686_WriteAC97Codec_sub(d0);
}

static void via686_channel_reset(void)
{
	int via_sound_iobase = g_pci.base0;

	outportb(via_sound_iobase + VIA_REG_OFFSET_CONTROL,
				VIA_REG_CTRL_PAUSE | VIA_REG_CTRL_TERMINATE | VIA_REG_CTRL_RESET);
	udelay(50);
	outportb(via_sound_iobase + VIA_REG_OFFSET_CONTROL, 0x00);
	outportb(via_sound_iobase + VIA_REG_OFFSET_STATUS, 0xFF);
	if(via686_device_type == DEVICE_VT82C686){
		outportb(via_sound_iobase + VIA_REG_OFFSET_TYPE, 0x00);
	}
	outportl(via_sound_iobase + VIA_REG_OFFSET_CURR_PTR, 0);
}

static DWORD via686_current_pos(void)
{
	DWORD d0 = 0;
	DWORD d1 = 0;
	DWORD d2;
	int limit = 1024;

	w_enter_critical();
	while(limit > 0){
		d0 = inportl(g_pci.base0 + VIA_REG_OFFSET_CURR_PTR);
		d1 = inportl(g_pci.base0 + VIA_REG_PLAYBACK_CURR_COUNT);
		d2 = inportl(g_pci.base0 + VIA_REG_OFFSET_CURR_PTR);
		if(d0 == d2) break;
		limit -= 1;
	}
	w_exit_critical();

	d0 = d0 - g_dosmem4k_phys_table[0];
	d0 = d0 >> 3;
	d0 = d0 - 1;
	d0 = (15 - d0) & 0x0F;
	d0 = (d0 * 4096) + (d1 & 0x00FFFFFF);
	d0 = d0 & 0x0000FFFF;

	return d0;
}

static void via686_exit(void)
{
	via686_channel_reset();
	if(vds_helper_unlock(get_address_dosmem64k_for_dma(), &g_dosmem64k_phys_table[0], 16) == FALSE){
		set_error_message("vds_helper_unlock error\n");
	}
	if(vds_helper_unlock(get_address_dosmem4k(), &g_dosmem4k_phys_table[0], 1) == FALSE){
		set_error_message("vds_helper_unlock error\n");
	}
	free_dosmem64k_for_dma();
	free_dosmem4k();
	wavedevice_struct_init();
}

static void via686_set_volume(void)
{
	BYTE vol;

	if(sound_device_master_volume == -1){
		vol = 0x08;    // default, -12dB
	}else{
		vol = sound_device_master_volume;
	}
	/* reset codec */
	via686_WriteAC97Codec(AC97_RESET, 0x00);
	udelay(100);

	via686_WriteAC97Codec(AC97_MASTER,	  vol * 0x0101);		/* Master Volume */
//	via686_WriteAC97Codec(AC97_HEADPHONE, vol * 0x0101);		/* Headphone Volume (optional) */
	via686_WriteAC97Codec(AC97_PCM, 0x0808);					/* PCM OUT	0dB */
}


static void set_irq_edge_or_level(int irq, int edge_or_level)
{
	/*
	 *	edge_or_level == 0	   set edge  triggered
	 *				  != 0	   set level triggered
	 */

	int bitmap;

	if(irq > 15) return;
	if(edge_or_level != 0) edge_or_level = 1;
	if(irq < 8){
		bitmap = inportb(0x04D0);
		bitmap = bitmap & ~(1 << irq);
		bitmap = bitmap | (edge_or_level << irq);
		outportb(0x04D0, bitmap);
	}else{
		irq -= 8;
		bitmap = inportb(0x04D1);
		bitmap = bitmap & ~(1 << irq);
		bitmap = bitmap | (edge_or_level << irq);
		outportb(0x04D1, bitmap);
	}
}


static void via686_chip_init(BOOL initialize)
{
	BYTE data;

	if(initialize == TRUE){
		if(via686_device_type == DEVICE_VT82C686){
			/* deassert ACLink reset, force SYNC */
			pci_write_config_byte(&g_pci, 0x41, 0xE0);
			udelay(100);
			pci_write_config_byte(&g_pci, 0x41, 0x00);
			udelay(100);
			/* ACLink on, deassert ACLink reset, VSR, SGD data out */
			/* note - FM data out has trouble with non VRA codecs !! */
			pci_write_config_byte(&g_pci, 0x41, 0xCD);
			udelay(100);

/********************************************************************
 * via686 soundblater init
 ********************************************************************/
			/*	enable sbpro, mpu401, opl and gameport	*/
			pci_write_config_byte(&g_pci, 0x42, 0x4F);
			/*	sbpro:220, mpu401:330, dma:1, irq:5  */
			pci_write_config_byte(&g_pci, 0x43, 0x1C);

			set_irq_edge_or_level(5, 0);

		}else{
			/* deassert ACLink reset */
			pci_write_config_byte(&g_pci, 0x41, 0xE0);
			udelay(100);
			/* deassert ACLink reset, force SYNC (warm AC'97 reset) */
			pci_write_config_byte(&g_pci, 0x41, 0x60);
			udelay(2);
			/* ACLink on, deassert ACLink reset, VSR, SGD data out */
			pci_write_config_byte(&g_pci, 0x41, 0xcc);
		}
		via686_set_volume();
	}else{
		if(0){
			if(via686_device_type == DEVICE_VT82C686){
				pci_read_config_byte(&g_pci, 0x41, &data);
				/* enable read channel pcm data output */
				data |= 0x04;
				pci_write_config_byte(&g_pci, 0x41, 0xCD);
			}
		}
	}
}


static BOOL via686_start(int rate, BOOL initialize)
{
	static char *device_name_via686 = NULL;
	DWORD data;
	int d0;
	DWORD a0;

	/* now via686 driver may work in dos-box under windows9x */
	if(0){
		if(detect_windows() == TRUE){
			set_error_message("VIA686: Windows detected. VIA686 driver only works in Pure DOS.\n");
			return FALSE;
		}
	}

	if(check_pci_bios() == FALSE){
		set_error_message("VIA686: No PCI BIOS found.\n");
		return FALSE;
	}

	/* detection */
	d0 = 0;
	while(via686_dev_list[d0].vender_id != 0x0000){
		g_pci.vender_id 	= via686_dev_list[d0].vender_id;
		g_pci.device_id 	= via686_dev_list[d0].device_id;
		g_pci.sub_vender_id = via686_dev_list[d0].sub_vender_id;
		g_pci.sub_device_id = via686_dev_list[d0].sub_device_id;
		if(find_pci_device(&g_pci) == TRUE){
			logerror_("VIA686: %s found.\n", via686_dev_list[d0].string);
			break;
		}
		d0 += 1;
	}
	if(via686_dev_list[d0].vender_id == 0x0000){
		set_error_message("VIA686: no compatble device found.\n");
		return FALSE;
	}
	via686_device_type = via686_dev_list[d0].type;
	device_name_via686 = via686_dev_list[d0].string;

	/* get io base address */
	pci_read_config_dword(&g_pci, 0x10, &data);
	data &= 0xFF00;
	if(data == 0){
		set_error_message("VIA686: device found, but disabled.\n");
		return FALSE;
	}
	g_pci.base0 = data;
	logerror_("VIA686: pci base0 at I/O %04X\n", g_pci.base0);

	/* initialize */
	via686_chip_init(initialize);

	/* playing buffer in dos memory */
	if(allocate_dosmem64k_for_dma(_16BITSTEREO) == FALSE){
		set_error_message("VIA686: allocate_dosmem64k_for_dma() error.\n");
		printf("error\n");
		return FALSE;
	}
	/* get physical address every 4KB */
	if(vds_helper_lock(get_address_dosmem64k_for_dma(), &g_dosmem64k_phys_table[0], 16) == FALSE){
		set_error_message("VIA686: vds_helper_lock() error.\n");
		return FALSE;
	}

	/* scatter/gather table  buffer in dos memory */
	if(allocate_dosmem4k() == FALSE){
		set_error_message("VIA686: allocate_dosmem4k() error.\n");
		return FALSE;
	}
	/* get physical address */
	if(vds_helper_lock(get_address_dosmem4k(), &g_dosmem4k_phys_table[0], 1) == FALSE){
		set_error_message("VIA686: vds_helper_lock() error.\n");
		return FALSE;
	}

	/* stop and reset if playing */
	via686_channel_reset();

	/* setup scatter/gather table for playback */
	_farsetsel(_dos_ds);
	a0 = get_address_dosmem4k();
	d0 = 0;
	while(d0 < 16){
		_farnspokel(a0	  , g_dosmem64k_phys_table[d0]);
		_farnspokel(a0 + 4, 0x00001000);		// 4096 bytes = 1 buffer
		if(d0 == 15){
			_farnspokel(a0 + 4, 0x80001000);	// set end point
		}
		a0 += 8;
		d0 += 1;
	}

	/* set scatter/gather table pointer */
	outportl(g_pci.base0 + VIA_REG_OFFSET_CURR_PTR, g_dosmem4k_phys_table[0]);

	if(via686_device_type == DEVICE_VT82C686){
		/* auto-start, 16bitpcm, stereo */
		outportb(g_pci.base0 + VIA_REG_OFFSET_TYPE, 0xB0);
	}else{
		outportl(g_pci.base0 + VIA_REG_OFFSET_STOP_IDX,
				VIA_REG_TYPE_16BIT | VIA_REG_TYPE_STEREO | 0xFF000000);
	}
	/* force playback rate 48000Hz */
	data = via686_ReadAC97Codec(AC97_EXTENDED_ID);
	if( (data & AC97_EA_VRA) != 0 ){
		/* if VBR supported, disable VBR, force 48000Hz */
		via686_WriteAC97Codec(AC97_EXTENDED_STATUS, 0x0000);
	}
//	  via686_WriteAC97Codec(AC97_PCM_FRONT_DAC_RATE, 0x80BB);
	udelay(100);

	/* start playing */
	if(via686_device_type == DEVICE_VT82C686){
		outportb(g_pci.base0 + VIA_REG_OFFSET_CONTROL, VIA_REG_CTRL_START);
	}else{
		outportb(g_pci.base0 + VIA_REG_OFFSET_CONTROL, VIA_REG_CTRL_START | VIA_REG_CTRL_AUTOSTART);
	}

	wd.device_name	   = device_name_via686;
	wd.playback_rate   = rate;
	wd.pcm_format	   = _16BITSTEREO;
	wd.device_exit	   = via686_exit;
	wd.pcm_upload	   = common_pcm_upload_func;
	wd.get_current_pos = via686_current_pos;
	wd.initialized = TRUE;

	return TRUE;
}

/********************************************************************
	Intel810 and derivations driver

 note:
	This driver was tested with i810(ICH), SiS7012, VIA686.
 ********************************************************************/


/* capture block */
#define ICH_REG_PI_BDBAR		0x00	/* dword - buffer descriptor list base address */
#define ICH_REG_PI_CIV			0x04	/* byte - current index value */
#define ICH_REG_PI_LVI			0x05	/* byte - last valid index */
#define   ICH_REG_LVI_MASK		0x1f
#define ICH_REG_PI_SR			0x06	/* byte - status register */
#define   ICH_FIFOE 		0x10	/* FIFO error */
#define   ICH_BCIS			0x08	/* buffer completion interrupt status */
#define   ICH_LVBCI 		0x04	/* last valid buffer completion interrupt */
#define   ICH_CELV			0x02	/* current equals last valid */
#define   ICH_DCH			0x01	/* DMA controller halted */
#define ICH_REG_PI_PICB 		0x08	/* word - position in current buffer */
#define ICH_REG_PI_PIV			0x0a	/* byte - prefetched index value */
#define   ICH_REG_PIV_MASK		0x1f	/* mask */
#define ICH_REG_PI_CR			0x0b	/* byte - control register */
#define   ICH_IOCE			0x10	/* interrupt on completion enable */
#define   ICH_FEIE			0x08	/* fifo error interrupt enable */
#define   ICH_LVBIE 		0x04	/* last valid buffer interrupt enable */
#define   ICH_RESETREGS 		0x02	/* reset busmaster registers */
#define   ICH_STARTBM			0x01	/* start busmaster operation */
/* playback block */
#define ICH_REG_PO_BDBAR		0x10	/* dword - buffer descriptor list base address */
#define ICH_REG_PO_CIV			0x14	/* byte - current index value */
#define ICH_REG_PO_LVI			0x15	/* byte - last valid command */
#define ICH_REG_PO_SR			0x16	/* byte - status register */
#define ICH_REG_PO_PICB 		0x18	/* word - position in current buffer */
#define ICH_REG_PO_PIV			0x1a	/* byte - prefetched index value */
#define ICH_REG_PO_CR			0x1b	/* byte - control register */
/* mic capture block */
#define ICH_REG_MC_BDBAR		0x20	/* dword - buffer descriptor list base address */
#define ICH_REG_MC_CIV			0x24	/* byte - current index value */
#define ICH_REG_MC_LVI			0x25	/* byte - last valid command */
#define ICH_REG_MC_SR			0x26	/* byte - status register */
#define ICH_REG_MC_PICB 		0x28	/* word - position in current buffer */
#define ICH_REG_MC_PIV			0x2a	/* byte - prefetched index value */
#define ICH_REG_MC_CR			0x2b	/* byte - control register */
/* global block */
#define ICH_REG_GLOB_CNT		0x2c	/* dword - global control */
#define   ICH_PCM_246_MASK	0x00300000	/* 6 channels (not all chips) */
#define   ICH_PCM_6 	0x00200000	/* 6 channels (not all chips) */
#define   ICH_PCM_4 	0x00100000	/* 4 channels (not all chips) */
#define   ICH_PCM_2 	0x00000000	/* 2 channels (stereo) */
#define   ICH_SRIE		0x00000020	/* secondary resume interrupt enable */
#define   ICH_PRIE		0x00000010	/* primary resume interrupt enable */
#define   ICH_ACLINK		0x00000008	/* AClink shut off */
#define   ICH_AC97WARM		0x00000004	/* AC'97 warm reset */
#define   ICH_AC97COLD		0x00000002	/* AC'97 cold reset */
#define   ICH_GIE		0x00000001	/* GPI interrupt enable */
#define ICH_REG_GLOB_STA		0x30	/* dword - global status */
#define   ICH_MD3		0x00020000	/* modem power down semaphore */
#define   ICH_AD3		0x00010000	/* audio power down semaphore */
#define   ICH_RCS		0x00008000	/* read completion status */
#define   ICH_BIT3		0x00004000	/* bit 3 slot 12 */
#define   ICH_BIT2		0x00002000	/* bit 2 slot 12 */
#define   ICH_BIT1		0x00001000	/* bit 1 slot 12 */
#define   ICH_SRI		0x00000800	/* secondary resume interrupt */
#define   ICH_PRI		0x00000400	/* primary resume interrupt */
#define   ICH_SCR		0x00000200	/* secondary codec ready */
#define   ICH_PCR		0x00000100	/* primary codec ready */
#define   ICH_MCINT 	0x00000080	/* MIC capture interrupt */
#define   ICH_POINT 	0x00000040	/* playback interrupt */
#define   ICH_PIINT 	0x00000020	/* capture interrupt */
#define   ICH_MOINT 	0x00000004	/* modem playback interrupt */
#define   ICH_MIINT 	0x00000002	/* modem capture interrupt */
#define   ICH_GSCI		0x00000001	/* GPI status change interrupt */
#define ICH_REG_ACC_SEMA		0x34	/* byte - codec write semaphore */
#define   ICH_CAS			0x01	/* codec access semaphore */

#define ICH_MAX_FRAGS		32		/* max hw frags */


#define DEVICE_INTEL	0
#define DEVICE_SIS		1

static AC97_DEVICE_LIST intel_ich_dev_list[] = {
	{ 0x8086, 0x2415, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "Intel 82801AA integrated AC97 audio" },
	{ 0x8086, 0x2425, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "Intel 82901AB integrated AC97 audio" },
	{ 0x8086, 0x2445, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "Intel 82801BA integrated AC97 audio" },
	{ 0x8086, 0x2485, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "Intel ICH3 integrated AC97 audio"	  },
	{ 0x8086, 0x24c5, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "Intel ICH4 integrated AC97 audio"	  },
	{ 0x8086, 0x7195, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "Intel 440MX integrated AC97 audio"   },
	{ 0x1039, 0x7012, PCI_ANY_ID, PCI_ANY_ID, DEVICE_SIS,	"SiS SI7012 AC97 audio" 			  },
	{ 0x10de, 0x01b1, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "NVIDIA NFORCE integrated AC97 audio" },
	{ 0x1022, 0x764d, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "AMD AMD8111 integrated AC97 audio"   },
	{ 0x1022, 0x7445, PCI_ANY_ID, PCI_ANY_ID, DEVICE_INTEL, "AMD AMD768 integrated AC97 audio"	  },

	{ 0x0000, 0x0000, PCI_ANY_ID, PCI_ANY_ID, 0, "" }
};

static int intel_ich_device_type = DEVICE_INTEL;


static BOOL intel_ich_codec_semaphore(void)
{
	DWORD d0;
	int limit = 1024;

	/* codec ready ? */
	d0 = inportl(g_pci.base1 + ICH_REG_GLOB_STA);
	if((d0 & ICH_PCR) == 0){
		return TRUE;
	}

	do{
		d0 = inportb(g_pci.base1 + ICH_REG_ACC_SEMA);
		if((d0 & ICH_CAS) == 0){
			return TRUE;
		}
		udelay(10);
	}while(limit-- > 0);

	return FALSE;
}

static WORD intel_ich_ReadAC97Codec(BYTE idx)
{
	intel_ich_codec_semaphore();
	return inportw(g_pci.base0 + idx);
}

static void intel_ich_WriteAC97Codec(BYTE idx, WORD data)
{
	intel_ich_codec_semaphore();
	outportw(g_pci.base0 + idx, data);
}

static void intel_ich_stop(void)
{
	outportb(g_pci.base1 + ICH_REG_PO_CR, 0x00);
	udelay(100);
	outportb(g_pci.base1 + ICH_REG_PO_CR, 0x02);
	udelay(100);
}

static void intel_ich_set_volume(void)
{
	BYTE vol;

	if(sound_device_master_volume == -1){
		vol = 0;	// default, 0dB
	}else{
		vol = sound_device_master_volume;
	}
	intel_ich_WriteAC97Codec(AC97_RESET, 0x00);

	intel_ich_WriteAC97Codec(AC97_MASTER,	 vol * 0x0101); 	/* Master Volume */
//	intel_ich_WriteAC97Codec(AC97_HEADPHONE, vol * 0x0101); 	/* Headphone Volume (optional) */
	intel_ich_WriteAC97Codec(AC97_PCM, 0x0808); 				/* PCM OUT	0dB */
}




static void intel_ich_init(void)
{
	DWORD cnt;
	int limit;

	/* put logic to right state */
	/* first clear status bits */
	cnt = inportl(g_pci.base1 + ICH_REG_GLOB_STA);
	outportl(g_pci.base1 + ICH_REG_GLOB_STA,
			cnt & (ICH_RCS | ICH_MCINT | ICH_POINT | ICH_PIINT));

	/* ACLink on, 2 channels */
	cnt = inportl(g_pci.base1 + ICH_REG_GLOB_CNT);
	cnt &= ~(ICH_ACLINK | ICH_PCM_246_MASK);
	/* finish cold or do warm reset */
	cnt |= (cnt & ICH_AC97COLD) == 0 ? ICH_AC97COLD : ICH_AC97WARM;
	outportl(g_pci.base1 + ICH_REG_GLOB_CNT, cnt);
	limit = 1024;
	do {
		if((inportl(g_pci.base1 + ICH_REG_GLOB_CNT) & ICH_AC97WARM) == 0) break;
		udelay(1000);
	} while (limit-- > 0);

	/* wait for primary codec ready status.
	 * Once it becomes ready it should remain ready
	 * as long as we do not disable the ac97 link.
	 */
	limit = 1024;
	do {
		if (inportl(g_pci.base1 + ICH_REG_GLOB_STA) & ICH_PCR) break;
		udelay(1000);
	} while (limit-- > 0);

	inportw(g_pci.base0);	 /* clear semaphore flag */

	/* disable interrupts */
	outportb(g_pci.base1 + ICH_REG_PI_CR, 0x00);
	outportb(g_pci.base1 + ICH_REG_PO_CR, 0x00);
	outportb(g_pci.base1 + ICH_REG_MC_CR, 0x00);
	/* reset channels */
	outportb(g_pci.base1 + ICH_REG_PI_CR, ICH_RESETREGS);
	outportb(g_pci.base1 + ICH_REG_PO_CR, ICH_RESETREGS);
	outportb(g_pci.base1 + ICH_REG_MC_CR, ICH_RESETREGS);

	intel_ich_set_volume();
}



static void intel_ich_exit(void)
{
	intel_ich_stop();
	if(vds_helper_unlock(get_address_dosmem64k_for_dma(), &g_dosmem64k_phys_table[0], 16) == FALSE){
		set_error_message("vds_helper_unlock error\n");
	}
	if(vds_helper_unlock(get_address_dosmem4k(), &g_dosmem4k_phys_table[0], 1) == FALSE){
		set_error_message("vds_helper_unlock error\n");
	}
        free_dosmem64k_for_dma();
        free_dosmem4k();
        wavedevice_struct_init();
}

static DWORD intel_ich_current_pos(void)
{
	BYTE d0 = 0;
	WORD d1 = 0;
	BYTE d2;
	DWORD d3;
	int limit = 1024;

	w_enter_critical();
	while(limit > 0){
		if(intel_ich_device_type == DEVICE_SIS){
			d0 = inportb(g_pci.base1 + ICH_REG_PO_CIV);
			d1 = inportw(g_pci.base1 + ICH_REG_PO_SR);
			d2 = inportb(g_pci.base1 + ICH_REG_PO_CIV);
		}else{
			d0 = inportb(g_pci.base1 + ICH_REG_PO_CIV);
			d1 = inportw(g_pci.base1 + ICH_REG_PO_PICB);
			d2 = inportb(g_pci.base1 + ICH_REG_PO_CIV);
		}
		if(d0 == d2) break;
		limit -= 1;
	}
	outportb(g_pci.base1 + ICH_REG_PO_LVI, (d0 + 31) & 0x1F);
	w_exit_critical();

	d3 = (15 - d0) & 0x0F;
	d3 = (d3 * 4096);
	if(intel_ich_device_type == DEVICE_SIS){
		d3 = d3 + d1;
	}else{
		d3 = d3 + (d1 << 1);
	}
	d3 = d3 & 0x0000FFFF;

//	  for debug
//	  d3 = (d0 << 16) | d1;

	return d3;
}


static BOOL intel_ich_start(int rate, BOOL initialize)
{
	static char *device_name_intel_ich = NULL;
	DWORD data;
	int d0;
	DWORD a0;

	/* now intel_ich driver may work in dos box under windows9x */
	if(0){
		if(detect_windows() == TRUE){
			set_error_message("Intel_ICH: Windows detected. Intel_ICH driver only works in Pure DOS.\n");
			return FALSE;
		}
	}

	if(check_pci_bios() == FALSE){
		set_error_message("Intel_ICH: No PCI BIOS found.\n");
		return FALSE;
	}

	/* detection */
	d0 = 0;
	while(intel_ich_dev_list[d0].vender_id != 0x0000){
		g_pci.vender_id 	= intel_ich_dev_list[d0].vender_id;
		g_pci.device_id 	= intel_ich_dev_list[d0].device_id;
		g_pci.sub_vender_id = intel_ich_dev_list[d0].sub_vender_id;
		g_pci.sub_device_id = intel_ich_dev_list[d0].sub_device_id;
		if(find_pci_device(&g_pci) == TRUE){
			logerror_("Intel_ICH: %s found.\n", intel_ich_dev_list[d0].string);
			break;
		}
		d0 += 1;
	}
	if(intel_ich_dev_list[d0].vender_id == 0x0000){
		set_error_message("Intel_ICH: no compatble device found.\n");
		return FALSE;
	}
	intel_ich_device_type = intel_ich_dev_list[d0].type;
	device_name_intel_ich = intel_ich_dev_list[d0].string;

	/* get io base address */
	pci_read_config_dword(&g_pci, 0x10, &data);
	data &= 0xFF00;
	if(data == 0){
		set_error_message("Intel_ICH: device found, but disabled.\n");
		return FALSE;
	}
	g_pci.base0 = data;
	logerror_("Intel_ICH: PCI BASE0 at I/O %04X\n", g_pci.base0);

	pci_read_config_dword(&g_pci, 0x14, &data);
	data &= 0xFFC0;
	if(data == 0){
		set_error_message("Intel_ICH: device found, but disabled.\n");
		return FALSE;
	}
	g_pci.base1 = data;
	logerror_("Intel_ICH: PCI BASE1 at I/O %04X\n", g_pci.base1);

	/* stop and reset if playing */
	intel_ich_stop();

	if(initialize == TRUE){
		/* initialize */
		intel_ich_init();
	}

	/* playing buffer in dos memory */
	if(allocate_dosmem64k_for_dma(_16BITSTEREO) == FALSE){
		set_error_message("Intel_ICH: allocate_dosmem64k_for_dma() error.\n");
		printf("error\n");
		return FALSE;
	}
	/* get physical address every 4KB */
	if(vds_helper_lock(get_address_dosmem64k_for_dma(), &g_dosmem64k_phys_table[0], 16) == FALSE){
		set_error_message("Intel_ICH: vds_helper_lock() error.\n");
		return FALSE;
	}

	/* scatter/gather table buffer in dos memory */
	if(allocate_dosmem4k() == FALSE){
		set_error_message("Intel_ICH: allocate_dosmem4k() error.\n");
		return FALSE;
	}
	/* get physical address */
	if(vds_helper_lock(get_address_dosmem4k(), &g_dosmem4k_phys_table[0], 1) == FALSE){
		set_error_message("Intel_ICH: vds_helper_lock() error.\n");
		return FALSE;
	}

	/* setup scatter/gather table for playback */
	_farsetsel(_dos_ds);
	a0 = get_address_dosmem4k();
	d0 = 0;
	while(d0 < 32){
		_farnspokel(a0	  , g_dosmem64k_phys_table[d0 & 0x0F]);
		if(intel_ich_device_type == DEVICE_SIS){
			_farnspokel(a0 + 4, 0x00001000);		// in bytes   (4096 bytes = 1 buffer)
		}else{
			_farnspokel(a0 + 4, 0x00000800);		// in samples (16bit pcm)
		}
		a0 += 8;
		d0 += 1;
	}

	/* set scatter/gather table pointer */
	outportl(g_pci.base1 + ICH_REG_PO_BDBAR, g_dosmem4k_phys_table[0]);
	outportb(g_pci.base1 + ICH_REG_PO_LVI, 0x1F);

	/* 16bitpcm, stereo */
	data = inportl(g_pci.base1 + ICH_REG_GLOB_CNT);
	data = data & ~ICH_PCM_246_MASK;
	data = data & ~ICH_GIE;
	outportl(g_pci.base1 + ICH_REG_GLOB_CNT, data);

	/* force playback rate 48000Hz */
	data = intel_ich_ReadAC97Codec(AC97_EXTENDED_ID);
	if( (data & AC97_EA_VRA) != 0 ){
		/* if VBR supported, disable VBR, force 48000Hz */
		intel_ich_WriteAC97Codec(AC97_EXTENDED_STATUS, 0x0000);
	}
	  intel_ich_WriteAC97Codec(AC97_PCM_FRONT_DAC_RATE, 0x80BB);
//       	  intel_ich_WriteAC97Codec(AC97_PCM_FRONT_DAC_RATE, 22500);
  //        data = intel_ich_ReadAC97Codec(AC97_PCM_FRONT_DAC_RATE);
	
	  udelay(100);

	/* start playing */
	outportb(g_pci.base1 + ICH_REG_PO_CR, ICH_STARTBM);

	wd.device_name	   = device_name_intel_ich;
	wd.playback_rate   = rate;
	wd.pcm_format	   = _16BITSTEREO;
	wd.device_exit	   = intel_ich_exit;
	wd.pcm_upload	   = common_pcm_upload_func;
	wd.get_current_pos = intel_ich_current_pos;
	wd.initialized = TRUE;

	return TRUE;
}

/********************************************************************
	AC97 auto-detection
 ********************************************************************/

static BOOL ac97_auto_detect_start(int rate, BOOL initialize)
{
	clear_error_message();
	if(via686_start(rate, initialize) == TRUE) return TRUE;

//	  clear_error_message();
	if(intel_ich_start(rate, initialize) == TRUE) return TRUE;

//	  clear_error_message();
	set_error_message("AC97_AUTODETECT: no supported device found.\n");
	return FALSE;
}

/********************************************************************
	"w_" sound system funtions
 ********************************************************************/

void w_set_watermark(float latency, DWORD samples_per_frame)
{
	if(wd.initialized == FALSE) return;

	if(latency < 1.0) latency = 1.0;
	g_latencym = latency;
	g_latency = samples_per_frame * latency;
	g_dma_average_cnt = samples_per_frame << DMA_AVERAGE_SHIFT_COUNT;
	g_samples_per_frame = samples_per_frame;
	w_reset_write_cursor(0);

	w_clear_buffer();
}


void w_set_master_volume(int volume)
{
	g_master_volume = volume;
}




DWORD w_get_current_req(void)
{
	if(wd.initialized == FALSE) return 0;
	return g_current_req;
}

DWORD w_get_next_req(void)
{
	if(wd.initialized == FALSE) return 0;
	return g_next_req;
}


static void calc_next_req(void)
{
	if(wd.initialized == FALSE) return;

	g_dma_remainder += g_dma_average_cnt;
	g_dma_remainder += g_dma_dt;
	g_dma_dt = 0;
	g_current_req = g_next_req;
	g_next_req = g_dma_remainder >> DMA_AVERAGE_SHIFT_COUNT;
	g_dma_remainder = g_dma_remainder & DMA_AVERAGE_MASK;

	if(g_current_req == 0 || g_current_req > SAMPLECNT) g_current_req = 1;
	if(g_next_req	 == 0 ||	g_next_req > SAMPLECNT) g_next_req = 1;
}


void w_reverse_stereo(int flag)
{
	stereo_reverse_flag = (flag) ? TRUE : FALSE;
}


void w_lock_mixing_buffer(int currentsamplecount)
{
	DWORD d0;

	if(wd.initialized == FALSE) return;

	if(currentsamplecount != 0)
		g_current_req = currentsamplecount;

	if(g_current_req > SAMPLECNT)
		g_current_req = SAMPLECNT;
	if(g_current_req == 0)
		g_current_req = 1;

	d0 = 0;
	while(d0 < g_current_req){
		mixing_buff[(d0*2) + 0] = 0;
		mixing_buff[(d0*2) + 1] = 0;
		d0 += 1;
	}
}


void w_mixing_zero(void)
{
	int d1;
	int leftindex  = 0;
	int rightindex = 1;

	if(stereo_reverse_flag == TRUE){
		leftindex  = 1;
		rightindex = 0;
	}

	if(wd.initialized == FALSE) return;
	if(g_current_req > SAMPLECNT) return;

	d1 = 0;
	while(d1 < g_current_req){
		mixing_buff[(d1*2) + leftindex ] = 0;
		mixing_buff[(d1*2) + rightindex] = 0;
		d1 += 1;
	}
}


void w_mixing8(char data[], DWORD length, int leftvol, int rightvol)
{
	DWORD dt;
	DWORD d0;
	int d1;
	long d2;
	int leftindex  = 0;
	int rightindex = 1;

	if(stereo_reverse_flag == TRUE){
		leftindex  = 1;
		rightindex = 0;
	}

	if(wd.initialized == FALSE) return;
	if(g_current_req > SAMPLECNT) return;

	dt = (length << 16) / g_current_req;
	d0 = 0;
	d1 = 0;
	while(d1 < g_current_req){
		d2 = (long)data[d0 >> 16];
		d2 = d2 << 8;
		mixing_buff[(d1*2) + leftindex ] += (d2 * leftvol)	/ 256;
		mixing_buff[(d1*2) + rightindex] += (d2 * rightvol) / 256;
		d0 += dt;
		d1 += 1;
	}
}

void w_mixing(short data[], DWORD length, int leftvol, int rightvol)
{
	DWORD dt;
	DWORD d0;
	int d1;
	long d2;
	int leftindex  = 0;
	int rightindex = 1;

	if(stereo_reverse_flag == TRUE){
		leftindex  = 1;
		rightindex = 0;
	}

	if(wd.initialized == FALSE) return;
	if(g_current_req > SAMPLECNT) return;

	dt = (length << 16) / g_current_req;
	d0 = 0;
	d1 = 0;
	while(d1 < g_current_req){
		d2 = (long)data[d0 >> 16];
		mixing_buff[(d1*2) + leftindex ] += (d2 * leftvol)	/ 256;
		mixing_buff[(d1*2) + rightindex] += (d2 * rightvol) / 256;
		d0 += dt;
		d1 += 1;
	}
}

void w_mixing_stereo(short data[], DWORD length, int leftvol, int rightvol)
{
	DWORD dt;
	DWORD d0;
	int d1;
	long left;
	long right;
	int leftindex  = 0;
	int rightindex = 1;

	if(stereo_reverse_flag == TRUE){
		leftindex  = 1;
		rightindex = 0;
	}

	if(wd.initialized == FALSE) return;
	if(g_current_req > SAMPLECNT) return;

	dt = (length << 16) / g_current_req;
	d0 = 0;
	d1 = 0;
	while(d1 < g_current_req){
		left  = (long)data[((d0 >> 16) * 2)];
		right = (long)data[((d0 >> 16) * 2) + 1];
		mixing_buff[(d1*2) + leftindex ] += (left  * leftvol)  / 256;
		mixing_buff[(d1*2) + rightindex] += (right * rightvol) / 256;
		d0 += dt;
		d1 += 1;
	}
}


static void common_pcm_upload_func(void)
{
	int d0;
	long d1;
	WORD d2;
	DWORD d3;
	DWORD start;
	DWORD end;

	start = -1;
	_farsetsel(_dos_ds);
	d0 = 0;
	while(d0 < g_current_req){
		g_write_cursor = g_write_cursor & SAMPLECNTMASK;
		if(wd.pcm_format == _16BITSTEREO){
			d1	= ((mixing_buff[(d0*2) + 0] * g_master_volume)/256);
			if(d1 >  32767) d1 =  32767;
			if(d1 < -32768) d1 = -32768;
			d3	= d1 & 0xFFFF;
			d1	= ((mixing_buff[(d0*2) + 1] * g_master_volume)/256);
			if(d1 >  32767) d1 =  32767;
			if(d1 < -32768) d1 = -32768;
			d3 |= d1 << 16;
			_farnspokel(g_wss_dma_addr + (g_write_cursor * 4), d3);
		}else if(wd.pcm_format == _8BITSTEREO){
			d1	= ((mixing_buff[(d0*2) + 0] * g_master_volume)/256);
			d2	= get_8bit_pcm_value(d1);
			d1	= ((mixing_buff[(d0*2) + 1] * g_master_volume)/256);
			d2 |= get_8bit_pcm_value(d1) << 8;
			_farnspokew(g_wss_dma_addr + (g_write_cursor * 2), d2);
		}else if(wd.pcm_format == _8BITMONO){
			d1	= ((mixing_buff[(d0*2) + 0] * g_master_volume)/256);
			d1 += ((mixing_buff[(d0*2) + 1] * g_master_volume)/256);
			d1	= d1 / 2;
			d2	= get_8bit_pcm_value(d1);
			_farnspokeb(g_wss_dma_addr + (g_write_cursor * 1), (BYTE)d2);
		}
		if(start == -1) start = g_write_cursor;
		g_write_cursor += 1;
		d0 += 1;
	}
	end = g_write_cursor;
}



void w_unlock_mixing_buffer(void)
{
	if(wd.pcm_upload != NULL){
		(*wd.pcm_upload)();
	}
}


static DWORD common_dma_current_pos(void)
{
	DWORD d0;
	w_enter_critical();
	d0 = _dma_todo(wd.isa_dma);
	w_exit_critical();

	return d0;
}




static DWORD _dma_counter(void)
{
	DWORD d0 = 0;

	if(wd.get_current_pos != NULL){
		d0 = (*wd.get_current_pos)();
	}
	return d0;
}



static DWORD get_played_sample_count(void)
{
	DWORD curr;
	DWORD d0;

	if(wd.initialized == FALSE) return 0;

	curr = w_get_play_cursor();
	d0 = w_calc_distance(curr, g_prev_play_cursor);
	g_prev_play_cursor = curr;

	return d0;
}





DWORD w_get_buffer_size(void)
{
	return SAMPLECNT;
}


DWORD w_get_play_cursor(void)
{
	return ((SAMPLECNT - 1) - (_dma_counter() >> g_dmacnt_shift_count)) & SAMPLECNTMASK;
}

DWORD w_get_write_cursor(void)
{
	return g_write_cursor;
}

void w_set_write_cursor(DWORD position)
{
	g_write_cursor = w_calc_position(position, 0);
}



void w_reset_watermark(void)
{
	w_reset_write_cursor(0);
}

void w_reset_watermark_ex(DWORD samplecount)
{
	w_reset_write_cursor(samplecount);
}

void w_reset_write_cursor(DWORD samplecount)
{
	DWORD current;

	if(wd.initialized == FALSE) return;

	current 		   = w_get_play_cursor();
	g_prev_play_cursor = current;

	if(samplecount <= g_latency){
		samplecount = g_latency - samplecount;
	}else{
		samplecount = 0;
	}
	w_set_write_cursor( w_calc_position(current, samplecount) );

	g_dma_remainder = 0;
	g_dma_dt = 0;
	calc_next_req();
	calc_next_req();
}


DWORD w_calc_distance(DWORD cursor1, DWORD cursor2)
{
	return (cursor1 - cursor2) & SAMPLECNTMASK;
}

DWORD w_calc_position(DWORD cursor1, DWORD cursor2)
{
	return (cursor1 + cursor2) & SAMPLECNTMASK;
}


DWORD w_get_requested_sample_count(void)
{
	DWORD d0;

	d0 = w_get_latency();
	if(d0 > w_get_buffer_size()/2){
		w_reset_watermark();
		return 0;
	}
	if(d0 >= g_latency) return 0;
	return g_latency - d0;
}


DWORD w_get_latency(void)
{
	DWORD d0;

	if(wd.initialized == FALSE){
		d0 = 0;
	}else{
		d0 = w_calc_distance(w_get_write_cursor(), w_get_play_cursor());
	}
	return d0;
}

int w_get_watermark_status(void)
{
	DWORD d0;
	int result;

	if(wd.initialized == FALSE) return 0;

	d0 = w_get_latency();

	if(d0 >= (DWORD)(g_latency*2.5)){
		result = W_WATERMARK_UNDERRUN;				// underrun or overrun
	}else if(d0 >= (DWORD)(g_latency*1.5)){
		result = W_WATERMARK_FULL;
	}else if(d0 >= (DWORD)g_latency){
		result = W_WATERMARK_GOOD;
	}else if(d0 >= g_samples_per_frame){
		result = W_WATERMARK_OK;
	}else if(d0 >= g_samples_per_frame/8){
		result = W_WATERMARK_ON_THE_EDGE;
	}else{
		result = W_WATERMARK_UNDERRUN;
	}

	return result;
}


static DWORD ofdt = 0;
static DWORD ufdt = 0;
static int dma_dt = 0;

DWORD w_adjust_latency_for_vsync(void)
{
	DWORD d0;

	if(wd.initialized == FALSE) return 0;

	d0 = w_calc_distance(w_get_write_cursor(), w_get_play_cursor());
	if( d0 >= (g_latency + 1) ){
		ofdt = d0 - g_latency;
		g_dma_overflow	+= 1;
		g_dma_underflow  = 0;
	}else if( d0 <= (g_latency - 1) ){
		ufdt = g_latency - d0;
		g_dma_overflow	 = 0;
		g_dma_underflow += 1;
	}else{
		g_dma_overflow	 = 0;
		g_dma_underflow  = 0;
	}
	if(g_dma_overflow > 16)
		g_dma_overflow = 16;
	if(g_dma_underflow > 16)
		g_dma_underflow = 16;
	ofdt = ofdt  * (g_dma_overflow	& 0xFE);
	ufdt = ufdt  * (g_dma_underflow & 0xFE);
	if(ofdt > (wd.playback_rate  >> 5))
		ofdt = (wd.playback_rate  >> 5);
	if(ufdt > (wd.playback_rate  >> 5))
		ufdt = (wd.playback_rate  >> 5);

	g_dma_dt = ufdt - ofdt;
	dma_dt = g_dma_dt;			// for debug

	calc_next_req();

	return d0;
}




float w_calc_samples_per_vsync(void (*__vsync)(void), int vsync_count)
{
	DWORD d0;
	DWORD result[2];
	int d1;
	int count;
	int limit;
	int d7;

	if(wd.initialized == FALSE) return 0;
	limit = 16;
	while(limit > 0){
		d7 = 0;
		while(d7 < 2){
			d1 = 0;
			while(d1 < 4){
				count = 0;
				while(count < vsync_count){
					(*__vsync)();
					count += 1;
				}
				get_played_sample_count();
				d1 += 1;
			}
			d0 = 0;
			d1 = 0;
			while(d1 < (256/8)){
				count = 0;
				while(count < vsync_count){
					(*__vsync)();
					count += 1;
				}
				d0 += get_played_sample_count();
				d1 += 1;
			}
			result[d7] = d0;
			d7 += 1;
		}
		if(result[0] > result[1]){
			d0 = result[0] - result[1];
		}else{
			d0 = result[1] - result[0];
		}
		if(d0 < 16) break;
		limit -= 1;
	}

	d0 = (result[0] + result[1]) * 4;
	g_latency = (d0 * g_latencym) / 256;
	g_dma_average_cnt = d0;
	calc_next_req();
	calc_next_req();

	return (float)d0/256;
}




void w_clear_buffer(void)
{
#if 0
	int d0;
	BYTE d1;

	if(wd.initialized == FALSE) return;
	d1 = 0x80;
	if(wd.pcm_format == _16BITSTEREO){
		d1 = 0x00;
	}
	_farsetsel(_dos_ds);
	d0 = 0;
	while(d0 < g_dma_buff_size){
		_farnspokeb(g_wss_dma_addr + d0, d1);
		d0 += 1;
	}

	if(DEVICE == DEVICE_GF1){
		u_set_volume(u_prim_voice + 0, 0);
		u_set_volume(u_prim_voice + 1, 0);

		d1 = 0x00;
		d0 = 0;
		while(d0 < 65536*2){
			u_upload(d0, d1);
			d0 += 1;
		}
		u_set_volume(u_prim_voice + 0, 0xFFFF);
		u_set_volume(u_prim_voice + 1, 0xFFFF);
	}
#else
	int d0;

	if(wd.initialized == FALSE) return;

	for(d0 = 0; d0 <= SAMPLECNT/1024; d0 += 1){
		w_lock_mixing_buffer(1024);
		w_mixing_zero();
		w_unlock_mixing_buffer();
	}
#endif
}





DWORD w_get_nominal_sample_rate(void)
{
	if(wd.initialized == FALSE) return 0;
	return wd.playback_rate;
}

char *w_get_device_name(void)
{
	if(wd.initialized == FALSE) return NULL;
	return wd.device_name;
}


void vga_vsync(void)
{
	w_enter_critical();
	while( (inportb(0x3da) & 0x08) != 0 );
	while( (inportb(0x3da) & 0x08) == 0 );
	w_exit_critical();
}


DWORD w_get_actual_sample_rate(void)
{
	DWORD d0;
	DWORD result[2];
	int d1;
	int d7;
	int limit;
	cycles_t prev;
	cycles_t curr;

	if(wd.initialized == FALSE) return 0;

	w_set_watermark(2.0, w_get_nominal_sample_rate()/60);

	prev = osd_cycles();
	while(1){
		if( (osd_cycles() - prev) >= osd_cycles_per_second()/2 ) break;
		get_played_sample_count();
	}

	limit = 16;
	while(limit > 0){
		d7 = 0;
		while(d7 < 2){
			d0 = 0;
			d1 = 65536;
			get_played_sample_count();
			while(d1 > 0){
				if(get_played_sample_count != 0) break;
				d1 -= 1;
			}
			prev = osd_cycles();
			while(1){
				d0 += get_played_sample_count();
				curr = osd_cycles();
				if( (curr - prev) >= osd_cycles_per_second()/2 ) break;
			}
			result[d7] = d0;
			d7 += 1;
		}
		if(result[0] > result[1]){
			d0 = result[0] - result[1];
		}else{
			d0 = result[1] - result[0];
		}
		if(d0 <= 8) break;
		limit -= 1;
	}
	actual_sample_rate = result[0] + result[1];
	return actual_sample_rate;
}


char *w_get_error_message(void)
{
	return get_error_message();
}

void w_set_device_master_volume(int volume)
{
	//		   max	  min
	// volume:	 0	-  31, or  -1(default)

	if( (0 <= volume) && (volume <= 31) ){
		sound_device_master_volume	= volume;
	}else{
		sound_device_master_volume	= -1;
	}
}


void w_sound_device_exit(void)
{
	w_clear_buffer();
	if(wd.device_exit != NULL){
		(*wd.device_exit)();
	}
	wd.initialized = FALSE;
}


int w_sound_device_init(int rate)
{
	int result;

	if(wd.initialized == TRUE){
		set_error_message("The sound device was already initialized.\n");
		return FALSE;
	}

	if(rate <= 0)	 rate = 22050;
	if(rate > 64000) rate = 64000;

        result = ac97_auto_detect_start(rate, TRUE);

	w_set_watermark(2.2, 1200);

	return result;
}




//////////////////////////////////////////////////////////////////////
// main() for debug



#ifdef _WSS_DEBUG


static double sin1dt;
static double sin2ldt;
static double sin2rdt;
static double sin1t;
static double sin1t2;
static double sin2lt;
static double sin2rt;


static void sin_stream_init(int rate)
{
	sin1dt	= 2.0 * M_PI * 880.0 / rate;
	sin2ldt = 2.0 * M_PI * 441.5 / rate;
	sin2rdt = 2.0 * M_PI * 438.5 / rate;
	sin1t	= 0;
	sin1t2	= 0;
	sin2lt	= 0;
	sin2rt	= 0;
}

static short sin_stream1(void)
{
	short d0;

	sin1t2 += 0.00005;
	if(sin1t2 > 2*M_PI) sin1t2 -= 2*M_PI;
	d0 = (short)(  16384 * sin(sin1t += (sin1dt + (sin1dt/4)*sin(sin1t2)))	);
	return d0;
}

static short sin_stream2l(void)
{
	short d0;

	d0 = (short)(16384 * sin(sin2lt += sin2ldt));
	return d0;
}

static short sin_stream2r(void)
{
	short d0;

	d0 = (short)(16384 * sin(sin2rt += sin2rdt));
	return d0;
}

#define NORMAL							 1
#define VSYNC							 2

#define FRAMES_PER_SEC	60

struct WavefileHead{
	unsigned char 	TAG[4];		//4
	unsigned int 	fileSize;	//8
	unsigned char 	waveTAG[4];	//12
	unsigned char 	ckID[4];	//16
	unsigned int 	nChuckSize;	//20
	short    int 	waveFormatTag;	//22
	short    int 	nChannel;	//24
	unsigned int 	nSamplePerSec; 	//28
	unsigned int 	nBytePerSec;	//32
	short    int 	nBlock;		//34
	short    int 	nBitSample;	//36
	unsigned char 	dataTag[4];	//40
	unsigned int 	dataSize;	//44
} wHead;
union BufT{
	        unsigned char ch[4];
		short int Value[2];
	}Buf;
unsigned char   filename[64];
unsigned char 	Wave[4];
int PlayorStop;

#define Wave_STATE_B16C2 1
#define Wave_STATE_B16C1 2
#define Wave_STATE_B08C2 3
#define Wave_STATE_B08C1 4

void playstop()
{
        printf("\nPress any key to continue.\n");
        PlayorStop=0;
        filename[0]='\0';
        getch();
}

int main(int argc, char *argv[])
{
	int d0,i;
	int actual_sample_rate;
//	short a0[65536];
	short int a1[65536*2];
        DWORD samples_this_frame = 0;
	DWORD samples_left_over;
	DWORD samples_per_frame;
        int mode;
	int Volume;
	int Wave_STATE;
	int TotalTime;
	int CurrentTime;
	int Time_Second;
	int Time_Minute;
        int OldTime;
        int BitStream;
       // char PlayIcon;
       // char count;
        FILE *wfile;
        mode = NORMAL;
        Volume=512;

        if (argc>1)
        {
        if (argv[1][0]=='/')
        {
          if (argv[1][1]=='?')
          {
            printf("AC97 Audio Wave Player\n\n");
            printf("WPLAY [filename.wav]\n");
            return;
          }
          else
          {
            printf ("Invalid option.\n");
            return;
          }
        }
        else
        {
          strcpy(filename,argv[1]);
        }
        }

       	if( w_sound_device_init(48000) == FALSE)
	{
		printf("%s\n", w_get_error_message() );
		return -1;
	}


        while(1)
        {
                clrscr();
	        puts("|==================================================|");
                puts("|               AC97 Audio Wave Player             |");
                puts("|                  2004-9-17  v1.13                |");
        	puts("|==================================================|");
                printf("Press +/- to change volume, <P> to pause, <Q> to quit, or other key to stop.\n\n");
                printf("Please enter a WAVE filename or <exit> to quit:\n");
                printf("File: ");
                if (filename[0]=='\0')
                  gets(filename);
                else
                  puts(filename);
        	PlayorStop=1;
                if((strcmp(filename,"exit")==NULL||strcmp(filename,"EXIT")==NULL))
                {
                        printf("\nBye! Have a nice day!\n");
                        break;
                }

        	if((wfile=fopen(filename,"rb"))==NULL)
		{
                        printf("Sorry! File not found.");
                        playstop();
                        continue;
		}
		fread(&wHead,sizeof(wHead),1,wfile);
		strcpy(Wave,"RIFF");
		for(i=0;i<4;i++)
		{
			if(wHead.TAG[i]!=Wave[i])
			{
                                printf("Wrong file type.");
                                playstop();
                                continue;
			}
		}
		strcpy(Wave,"WAVE");
		for(i=0;i<4;i++)
		{
			if(wHead.waveTAG[i]!=Wave[i])
			{
				for(i=0;i<4;i++)
				{
					putchar(wHead.waveTAG[i]);
				}
                                printf("Only supports WAVE type.");
                                playstop();
                                continue;
			}
		}
		strcpy(Wave,"fmt ");
		for(i=0;i<4;i++)
		{
			if(wHead.ckID[i]!=Wave[i])
			{
				
                                printf("Unsupported file format.");
                                playstop();
                                continue;
			}
		}

		if(wHead.waveFormatTag!=1)
		{
                        printf("Only supports PCM.");
                        playstop();
                        continue;
		}

                printf("\nSound device : %s\n\n", w_get_device_name() );
                printf("Wave file name = %s\n",filename);
                printf("Wave file size = %x\n",wHead.fileSize);
		printf("Data size      = %x\n",wHead.dataSize);
		printf("Channel        = %d\n",wHead.nChannel);
                printf("Sample percent = %d\n",wHead.nSamplePerSec);
                printf("Sample bits    = %d\n",wHead.nBitSample);
                printf("\nStarting to play..\n");

		TotalTime=wHead.fileSize/wHead.nSamplePerSec;
		TotalTime=TotalTime/wHead.nChannel;
		TotalTime=(TotalTime*8)/(wHead.nBitSample);
		Time_Second=TotalTime%60;
		Time_Minute=TotalTime/60;
		gotoxy(50,20);
                if(Time_Second<10)
                {
		        printf("%d : 0%d\n",Time_Minute,Time_Second);
                }
                else
                {
                        printf("%d : %d\n",Time_Minute,Time_Second);
		}
		gotoxy(50,21);
                printf("Volume: %d",Volume/64);
		Time_Minute=0;
		Time_Second=0;
		gotoxy(60,20);
		printf(" %d : 0%d",Time_Minute,Time_Second);
                             	
                if((wHead.nChannel==2)&&(wHead.nBitSample==16))
	       {
	       		Wave_STATE=Wave_STATE_B16C2;
	       }
	       else if((wHead.nChannel==1)&&(wHead.nBitSample==16))
	       {
	       		Wave_STATE=Wave_STATE_B16C1;
	       }
	        else if((wHead.nChannel==2)&&(wHead.nBitSample==8))
	       {
	       		Wave_STATE=Wave_STATE_B08C2;
	       }
	       else if((wHead.nChannel==1)&&(wHead.nBitSample==8))
	       {
	      	 	Wave_STATE=Wave_STATE_B08C1;
	       }
	       else
	       {
			Wave_STATE=0;
	       }


		wd.playback_rate   = wHead.nSamplePerSec;
                intel_ich_WriteAC97Codec(AC97_PCM_FRONT_DAC_RATE, wd.playback_rate);
//               printf("Nominal rate : %d Hz\n", w_get_nominal_sample_rate());
	
//               actual_sample_rate = w_get_actual_sample_rate();
//               printf("Actual rate  : %d Hz\n", actual_sample_rate);
               actual_sample_rate = wHead.nSamplePerSec;
		BitStream=48000/actual_sample_rate;
		gotoxy(70,21);
		printf(" %d Hz\n", actual_sample_rate);
		samples_this_frame = actual_sample_rate / FRAMES_PER_SEC;
		w_set_watermark(2.2, samples_this_frame);
		if(mode == VSYNC){
			samples_this_frame = (int)w_calc_samples_per_vsync(vga_vsync, 1);
			printf("samples per vsync : %d\n", samples_this_frame);
		}
		if(mode == NORMAL){
	
			samples_per_frame = (actual_sample_rate << 12) / FRAMES_PER_SEC;
			samples_left_over = 0;
		}
		/* main loop */

                gotoxy(75,20);		puts("->");
		while(PlayorStop)
		{
                /*
                    if(count>5)
                         {
			        switch(PlayIcon%4)
                                {
                                        case 0:  gotoxy(75,20);puts("||");break;
                                        case 1:  gotoxy(75,20);puts("//");break;
                                        case 2:  gotoxy(75,20);puts("--");break;
                                        case 3:  gotoxy(75,20);puts("\\");break;
                                 }
                         PlayIcon++;
                         count=0;
                         }
                         count++;
		*/
			if(mode == NORMAL){
				/* normal mode */
				samples_left_over += samples_per_frame;
				samples_this_frame = (DWORD)(samples_left_over >> 12);
				samples_left_over  = samples_left_over & 0xFFF;
			}else{
				/* vsync mode */
				samples_this_frame = w_get_next_req();
			}
	
	
			 /* do something here. start */
                      /* do something here. end */
                         CurrentTime=ftell(wfile);
                         CurrentTime=CurrentTime/wHead.nSamplePerSec;
		         CurrentTime=CurrentTime/wHead.nChannel;
		         CurrentTime=(CurrentTime*8)/(wHead.nBitSample);
                         if(CurrentTime!=OldTime)
                         {
                         	OldTime=CurrentTime;
		         	Time_Second=CurrentTime%60;
		         	Time_Minute=CurrentTime/60;
                         	gotoxy(60,20);
                            	if(Time_Second<10)
                            	{
		                        printf(" %d : 0%d\n",Time_Minute,Time_Second);
                             	}
                             	else
                             	{
                                        printf(" %d : %d\n",Time_Minute,Time_Second);
                             	}
                
                        }
			if(Wave_STATE==Wave_STATE_B16C2)
			{
		                for(d0 = 0;d0 < samples_this_frame/BitStream; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,4,1,wfile))
                                        {
                                                for(i=0;i<BitStream;i++)
                                                {
						        a1[(d0 * (BitStream)*2) + 0+i] = Buf.Value[0];
						        a1[(d0 * (BitStream)*2) + 1+i] = Buf.Value[1];

                                                }
					}
					else
					{     
                                                //  printf("File End\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
                                //i=BitStream;
                                /*
                                while(i>1)
                                {
                                        for(d0 = 0;d0 < samples_this_frame/i-1; d0 += 1)
                                        {
                                                a1[((d0+1)* 2) + 0] = (a1[((d0) * 2) + 0] + a1[((d0+2) * 2) + 0])/2;
                                                a1[((d0+1)* 2) + 1] = (a1[((d0) * 2) + 1] + a1[((d0+2) * 2) + 1])/2;
                                        }
                                        i--;
                                }
                                */
			}
			else if(Wave_STATE==Wave_STATE_B16C1)
			{
                        /*
				for(d0 = 0;d0 < samples_this_frame; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,2,1,wfile))
		        		{
						a1[(d0 * 2) + 0] = Buf.Value[0];
						a1[(d0 * 2) + 1] = Buf.Value[0];
					}
					else
					{    
						 //  printf("File end\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
                                */
                                for(d0 = 0;d0 < samples_this_frame/BitStream; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,2,1,wfile))
                                        {
                                                for(i=0;i<BitStream;i++)
                                                {
						        a1[(d0 * (BitStream)*2) + 0+i] = Buf.Value[0];
						        a1[(d0 * (BitStream)*2) + 1+i] = Buf.Value[0];

                                                }
					}
					else
					{     
						//  printf("File end\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
			}
			else if(Wave_STATE==Wave_STATE_B08C2)
			{
                        /*
			        for(d0 = 0;d0 < samples_this_frame; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,2,1,wfile))
		        		{
                                        	a1[(d0 * 2) + 0] =   (Buf.ch[0]);
                                                a1[(d0 * 2) + 0] =   a1[(d0 * 2) + 0]*16;
						a1[(d0 * 2) + 1] =   (Buf.ch[1]);
                                                a1[(d0 * 2) + 1] =   a1[(d0 * 2) + 1]*16;
					}
					else
					{    
						 //  printf("File end\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
                                */
                                for(d0 = 0;d0 < samples_this_frame/BitStream; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,2,1,wfile))
                                        {
                                                for(i=0;i<BitStream;i++)
                                                {
						        a1[(d0 * (BitStream)*2) + 0+i] = Buf.ch[0]*16;
						        a1[(d0 * (BitStream)*2) + 1+i] = Buf.ch[1]*16;

                                                }
					}
					else
					{     
						//  printf("File end\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
			}
			else if(Wave_STATE==Wave_STATE_B08C1)
			{
                        /*
				for(d0 = 0;d0 < samples_this_frame; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,1,1,wfile))
		        		{
						a1[(d0 * 2) + 0] = Buf.ch[0];
                                                a1[(d0 * 2) + 0] =   a1[(d0 * 2) + 0]*16;
						a1[(d0 * 2) + 1] =   (Buf.ch[0]);
                                                a1[(d0 * 2) + 1] =   a1[(d0 * 2) + 1]*16;
					}
					else
					{    
						//  printf("File end\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
                       */
                       for(d0 = 0;d0 < samples_this_frame/BitStream; d0 += 1)
		                {
		
		                       	if(fread(Buf.ch,1,1,wfile))
                                        {
                                                for(i=0;i<BitStream;i++)
                                                {
						        a1[(d0 * (BitStream)*2) + 0+i] = Buf.ch[0]*16;
						        a1[(d0 * (BitStream)*2) + 1+i] = Buf.ch[0]*16;

                                                }
					}
					else
					{     
					       //  printf("File end\n");
		                                fseek(wfile,44,SEEK_SET);
		                                break;
					}
		       
				}
			}
			else
			{
                                printf("Not supports this WAVE file\n");
				break;
			}
	
			if(mode == NORMAL)
			{
				/* normal mode */
				while(1){
					d0 = w_get_watermark_status();
					if(d0 == W_WATERMARK_GOOD || d0 == W_WATERMARK_OK || d0 == W_WATERMARK_ON_THE_EDGE) break;
					if(d0 == W_WATERMARK_UNDERRUN) {
						w_reset_watermark();
						//printf("\n reset watermark \n\n");
						break;
					}
				}
				//printf("latency:%5d, play:%5d, write:%5d \n",w_get_latency(),w_get_play_cursor(), w_get_write_cursor());
				w_lock_mixing_buffer(samples_this_frame);
			//	w_mixing(a0, samples_this_frame, 0, 256);
				//if(Wave_STATE==Wave_STATE_B16C2)
				//{
					w_mixing_stereo(a1, samples_this_frame, Volume, Volume);
				//}
				w_unlock_mixing_buffer();
			}else{
				/* vsync mode */
				w_lock_mixing_buffer(samples_this_frame);
				//w_mixing(a0, samples_this_frame, 0, 256);
				w_mixing_stereo(a1, samples_this_frame, Volume, Volume);
				w_unlock_mixing_buffer();
	
				vga_vsync();
	
				d0 = w_get_watermark_status();
				if(d0 == W_WATERMARK_ON_THE_EDGE || d0 == W_WATERMARK_UNDERRUN){
					w_reset_watermark();
			//		printf("\n reset watermark \n\n");
				}
				w_adjust_latency_for_vsync();
			//	printf("latency:%5d, play:%5d, write:%5d \n",
	//						w_get_latency(),
	//						w_get_play_cursor(), w_get_write_cursor());
			//	printf("over %5d, under %5d, ofdt %5d, ufdt %5d, dt %5d\n",
	//				g_dma_overflow, g_dma_underflow, ofdt, ufdt, dma_dt);
			}
	
			if( kbhit() )
			{

		                switch(getch())
	                        {               case 'p':
                                                case 'P':       w_clear_buffer();       gotoxy(75,20);		puts("||");		while(!kbhit());		getch();	gotoxy(75,20);		puts("->");	break;
                                                case 'Q':
                                                case 'q':       return;
					        case '+':	if(Volume<1024)		Volume+=64; 		gotoxy(50,21);		puts("       ");                gotoxy(50,21);   printf("Vol: %d",Volume/64);	break;
					        case '-':	if(Volume>0)		Volume-=64;		gotoxy(50,21);		puts("       ");                gotoxy(50,21);  printf("Vol: %d",Volume/64);	break;

                                                default:        PlayorStop=0;filename[0]='\0';break;
	                        }

	                }
		}


      		printf("\nFinished.\n");
//              w_sound_device_exit();
		w_clear_buffer();
        	fclose(wfile);
	}
                   w_sound_device_exit();

	return 0;
}


#endif
